/*
 * Created on Jun 23, 2003
 *
 */
package org.getopt.luke;

import java.awt.Color;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.math.BigInteger;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;

import javax.swing.JFileChooser;
import javax.swing.UIManager;

import org.apache.lucene.LucenePackage;
import org.apache.lucene.analysis.*;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexGate;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.TermDocs;
import org.apache.lucene.index.TermEnum;
import org.apache.lucene.index.TermFreqVector;
import org.apache.lucene.index.TermPositions;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.BooleanClause;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.Explanation;
import org.apache.lucene.search.FilteredTermEnum;
import org.apache.lucene.search.FuzzyQuery;
import org.apache.lucene.search.HitCollector;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.MultiPhraseQuery;
import org.apache.lucene.search.PhraseQuery;
import org.apache.lucene.search.PrefixQuery;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.RangeQuery;
import org.apache.lucene.search.Similarity;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.WildcardQuery;
import org.apache.lucene.search.spans.SpanQuery;
import org.apache.lucene.search.spans.Spans;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.IndexOutput;
import org.apache.lucene.store.MMapDirectory;
import org.getopt.luke.plugins.ScriptingPlugin;

import thinlet.FrameLauncher;
import thinlet.Thinlet;

/**
 * This class allows you to browse a <a href="jakarta.apache.org/lucene">Lucene
 * </a> index in several ways - by document, by term, by query, and by most
 * frequent terms.
 * 
 * @author Andrzej Bialecki <ab@getopt.org>
 *  
 */
public class Luke extends Thinlet implements ClipboardOwner {

  private Directory dir = null;
  private String pName = null;
  private IndexReader ir = null;
  private IndexSearcher is = null;
  private Collection fn = null;
  private String[] idxFields = null;
  private List plugins = new ArrayList();
  private Object error = null;
  private Object statmsg = null;
  private Analyzer stdAnalyzer = new StandardAnalyzer();
  private Analyzer analyzer = null;
  private QueryParser qp = null;
  private boolean readOnly = false;
  private Similarity similarity = null;
  
  /** Default salmon theme. */
  public static final int THEME_DEFAULT     = 0;
  /** Gray theme. */
  public static final int THEME_GRAY        = 1;
  /** Sandstone theme. */
  public static final int THEME_SANDSTONE   = 2;
  /** Sky blue theme. */
  public static final int THEME_SKY         = 3;
  /** Navy blue reverse theme. */
  public static final int THEME_NAVY        = 4;
  
  /** Theme color contants. */
  public int[][] themes = {
          {0xece9d0, 0x000000, 0xf5f4f0, 0x919b9a, 0xb0b0b0, 0xededed, 0xb9b9b9, 0xff8080, 0xc5c5dd}, // default
          {0xe6e6e6, 0x000000, 0xffffff, 0x909090, 0xb0b0b0, 0xededed, 0xb9b9b9, 0x89899a, 0xc5c5dd}, // gray
          {0xeeeecc, 0x000000, 0xffffff, 0x999966, 0xb0b096, 0xededcb, 0xcccc99, 0xcc6600, 0xffcc66}, // sandstone
          {0xf0f0ff, 0x0000a0, 0xffffff, 0x8080ff, 0xb0b0b0, 0xededed, 0xb0b0ff, 0xff0000, 0xfde0e0}, // sky
          {0x6375d6, 0xffffff, 0x7f8fdd, 0xd6dff5, 0x9caae5, 0x666666, 0x003399, 0xff3333, 0x666666}  // navy
  };

  private int numTerms = 0;
  private static boolean exitOnDestroy = false;
  private Class[] analyzers = null;

  private String baseDir = null;

  private Class[] defaultAnalyzers = { SimpleAnalyzer.class, StandardAnalyzer.class, StopAnalyzer.class,
      WhitespaceAnalyzer.class };

  private static final String MSG_NOINDEX = "FAILED: No index, or index is closed. Reopen it.";
  private static final String MSG_READONLY = "FAILED: Read-Only index.";

  /** Default constructor, loads preferences, initializes plugins and GUI. */ 
  public Luke() {
    super();
    Prefs.load();
    try {
      UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
    } catch (Exception e) {}
    setTheme(Prefs.getInteger(Prefs.P_THEME, THEME_DEFAULT));
    String fontName = Prefs.getProperty(Prefs.P_FONT_NAME, "SansSerif");
    String fontSize = Prefs.getProperty(Prefs.P_FONT_SIZE, "12.0");
    float fsize = 12.0f;
    try {
      fsize = Float.parseFloat(fontSize);
    } catch (Exception e) {};
    Font f = new Font(fontName, Font.PLAIN, (int)fsize);
    setFont(f);
    addComponent(this, "/xml/luke.xml", null, null);
    error = addComponent(null, "/xml/error.xml", null, null);
    statmsg = find("statmsg");
    // populate analyzers
    try {
      analyzers = ClassFinder.getInstantiableSubclasses(Analyzer.class);
      if (analyzers == null || analyzers.length == 0) analyzers = defaultAnalyzers;
      Object cbType = find("cbType");
      populateAnalyzers(cbType);
    } catch (Exception e) {
      e.printStackTrace();
    }
    loadPlugins();
    
  }

  /**
   * Set color theme for the UI.
   * @param which one of the predefined themes. For custom themes use {@link Thinlet#setColors(int, int, int, int, int, int, int, int, int)}.
   */
  public void setTheme(int which) {
    if (which < 0 || which >= themes.length) which = THEME_DEFAULT;
    int[] t = themes[which];
    setColors(t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8]);
    Prefs.setProperty(Prefs.P_THEME, which + "");
  }
  
  /**
   * Action handler to select color theme.
   * @param menu
   */
  public void actionTheme(Object menu) {
    String which = (String)getProperty(menu, "t");
    int t = THEME_DEFAULT;
    try {
      t = Integer.parseInt(which);
    } catch (Exception e) {};
    setTheme(t);
  }
  
  /**
   * Populate a combobox with the current list of analyzers.
   * @param combo
   */
  public void populateAnalyzers(Object combo) {
    removeAll(combo);
    String[] aNames = new String[analyzers.length];
    for (int i = 0; i < analyzers.length; i++) {
      aNames[i] = analyzers[i].getName();
    }
    Arrays.sort(aNames);
    for (int i = 0; i < aNames.length; i++) {
      Object choice = create("choice");
      setString(choice, "text", aNames[i]);
      add(combo, choice);
      if (i == 0) {
        setString(combo, "text", aNames[i]);
      }
    }
    setInteger(combo, "selected", 0);
  }

  /**
   * Return an array of available Analyzer implementations.
   * @return
   */
  public Class[] getAnalyzers() {
    return analyzers;
  }

  /**
   * Loads plugins. Plugins are first searched from the CLASSPATH, and then from a
   * plugin list contained in a resource file "/.plugins". The "/.plugins" resource file
   * has a simple format - one fully qualified class name per line. Blank lines and
   * lines starting with '#' are ignored.
   */
  private void loadPlugins() {
    List pluginClasses = new ArrayList();
    // try to find all plugins
    try {
      Class classes[] = ClassFinder.getInstantiableSubclasses(LukePlugin.class);
      if (classes != null && classes.length > 0) {
        pluginClasses.addAll(Arrays.asList(classes));
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    // load plugins declared in the ".plugins" file
    try {
      InputStream is = getClass().getResourceAsStream("/.plugins");
      if (is != null) {
        BufferedReader br = new BufferedReader(new InputStreamReader(is));
        String line = null;
        while ((line = br.readLine()) != null) {
          if (line.startsWith("#")) continue;
          if (line.trim().equals("")) continue;
          try {
            Class clazz = Class.forName(line.trim());
            if (clazz.getSuperclass().equals(LukePlugin.class) && !pluginClasses.contains(clazz)) {
              pluginClasses.add(clazz);
            }
          } catch (Throwable x) {
            //
          }
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    try {
      StringBuffer errors = new StringBuffer("Unable to load some plugins:");
      boolean failures = false;
      for (int i = 0; i < pluginClasses.size(); i++) {
        try {
          LukePlugin plugin = (LukePlugin) ((Class) pluginClasses.get(i)).getConstructor(new Class[0]).newInstance(
                  new Object[0]);
          String xul = plugin.getXULName();
          if (xul == null) continue;
          Object ui = parse(xul, plugin);
          plugin.setApplication(this);
          plugin.setMyUi(ui);
          plugins.add(plugin);
        } catch (Exception e) {
          failures = true;
          e.printStackTrace();
          errors.append("\n" + pluginClasses.get(i).toString());
        }
      }
      if (failures) {
        errorMsg(errors.toString());
      }
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg(e.toString());
    }
    if (plugins.size() == 0) return;
    initPlugins();
  }

  /**
   * Create UI for a single plugin.
   * @param tabs parent tabbedpane
   * @param plugin plugin instance
   */
  private void addPluginTab(Object tabs, LukePlugin plugin) {
    Object tab = create("tab");
    setColor(tab, "foreground", new Color(0x006000));
    setString(tab, "text", plugin.getPluginName());
    setFont(tab, getFont().deriveFont(Font.BOLD));
    add(tabs, tab);
    Object panel = create("panel");
    setInteger(panel, "gap", 2);
    setInteger(panel, "weightx", 1);
    setInteger(panel, "weighty", 1);
    setChoice(panel, "halign", "fill");
    setChoice(panel, "valign", "fill");
    setInteger(panel, "columns", 1);
    add(tab, panel);
    Object infobar = create("panel");
    setInteger(infobar, "gap", 8);
    setInteger(infobar, "top", 2);
    setInteger(infobar, "bottom", 2);
    setInteger(infobar, "weightx", 1);
    setChoice(infobar, "halign", "fill");
    setColor(infobar, "background", new Color(0xc0f0c0));
    add(panel, infobar);
    Object label = create("label");
    setString(label, "text", plugin.getPluginInfo());
    add(infobar, label);
    Object link = create("button");
    setChoice(link, "type", "link");
    setString(link, "text", plugin.getPluginHome());
    putProperty(link, "url", plugin.getPluginHome());
    setMethod(link, "action", "goUrl(this)", infobar, this);
    add(infobar, link);
    add(panel, create("separator"));
    add(panel, plugin.getMyUi());
  }

  /**
   * Return the list of active plugin instances.
   * @return
   */
  public List getPlugins() {
    return Collections.unmodifiableList(plugins);
  }
  
  /**
   * Get an already instantiated plugin, or null if such plugin was
   * not loaded on startup.
   * @param className fully qualified plugin classname
   * @return
   */
  public LukePlugin getPlugin(String className) {
    for (int i = 0; i < plugins.size(); i++) {
      Object plugin = plugins.get(i);
      if (plugin.getClass().getName().equals(className))
        return (LukePlugin)plugin;
    }
    return null;
  }
  
  /**
   * Display a message on the status bar for 5 seconds.
   * @param msg message to display. Too long messages will be truncated by the UI.
   */
  public void showStatus(final String msg) {
    Thread thr = new Thread() {
      public void run() {
        setString(statmsg, "text", msg);
        try {
          sleep(5000);
        } catch (Exception e) {}
        setString(statmsg, "text", "");
      }
    };
    thr.start();
  }

  /**
   * Add a Thinlet component from XUL file.
   * @param parent add the new component to this parent 
   * @param compView path to the XUL resource
   * @param handlerStr fully qualified classname of the handler to instantiate,
   * or null if the current class will become the handler
   * @param argv if not null, these arguments will be passed to the
   * appropriate constructor.
   * @return
   */
  public Object addComponent(Object parent, String compView, String handlerStr, Object[] argv) {
    Object res = null;
    Object handler = null;
    try {
      if (handlerStr != null) {
        if (argv == null) {
          handler = Class.forName(handlerStr).getConstructor(new Class[] { Thinlet.class }).newInstance(
                  new Object[] { this });
        } else {
          handler = Class.forName(handlerStr).getConstructor(new Class[] { Thinlet.class, Object[].class })
                  .newInstance(new Object[] { this, argv });
        }
      }
      if (handler != null) {
        res = parse(compView, handler);
      } else res = parse(compView);
      if (parent != null) {
        if (parent instanceof Thinlet)
          add(res);
        else add(parent, res);
      }
      return res;
    } catch (Exception exc) {
      exc.printStackTrace();
      errorMsg(exc.getMessage());
      return null;
    }
  }

  /**
   * Show a modal error dialog with OK button.
   * @param msg error message
   */
  public void errorMsg(String msg) {
    System.out.println(msg);
    Object fMsg = find(error, "msg");
    setString(fMsg, "text", msg);
    add(error);
  }

  /**
   * Show an "Open Index" dialog.
   *
   */
  public void actionOpen() {
    Object dialog = addComponent(this, "/xml/lukeinit.xml", null, null);
    Object path = find(dialog, "path");
    if (this.baseDir != null)
      setString(path, "text", this.baseDir);
    else setString(path, "text", System.getProperty("user.dir"));
  }

  /**
   * Browse for a directory, and put the selection result in the
   * indicated widget.
   * @param path Thinlet widget to put the result
   */
  public void openBrowse(Object path) {
    JFileChooser fd = new JFileChooser();
    fd.setDialogType(JFileChooser.OPEN_DIALOG);
    fd.setDialogTitle("Select Index directory");
    fd.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
    fd.setFileHidingEnabled(false);
    String strPath = getString(path, "text");
    if (strPath != null) strPath.trim();
    if (strPath != null && strPath.length() > 0)
      fd.setCurrentDirectory(new File(strPath));
    else if (this.baseDir != null)
      fd.setCurrentDirectory(new File(this.baseDir));
    else fd.setCurrentDirectory(new File(System.getProperty("user.dir")));
    int res = fd.showOpenDialog(this);
    File iDir = null;
    if (res == JFileChooser.APPROVE_OPTION) iDir = fd.getSelectedFile();
    if (iDir != null && iDir.exists()) {
      if (!iDir.isDirectory()) iDir = iDir.getParentFile();
      setString(path, "text", iDir.toString());
    }
  }

  
  /**
   * Initialize MRU list of indexes in the open index dialog.
   * @param dialog
   */
  public void setupInit(Object dialog) {
    Object path = find(dialog, "path");
    syncMRU(path);
  }

  /**
   * Attempt to load the index with parameters specified in the dialog.
   * <p>NOTE: this method is invoked from the UI. If you need to open an index
   * programmatically, you should use {@link #openIndex(String, boolean, boolean, boolean)} instead.</p>
   * @param dialog UI dialog with parameters
   */
  public void openOk(Object dialog) {
    Object path = find(dialog, "path");
    pName = getString(path, "text").trim();

    boolean force = getBoolean(find(dialog, "force"), "selected");
    boolean mmap = getBoolean(find(dialog, "mmap"), "selected");
    if (pName == null || pName.trim().equals("") || !IndexReader.indexExists(pName)) {
      errorMsg("Invalid path, or not a Lucene index.");
      return;
    }
    readOnly = getBoolean(find(dialog, "ro"), "selected");
    Prefs.addToMruList(pName);
    syncMRU(path);
    remove(dialog);
    openIndex(pName, force, mmap, readOnly);
  }
  
  public void actionClose() {
    if (ir != null) {
      try {
        is.close();
        ir.close();
        dir.close();
      } catch (Exception e) {
        e.printStackTrace();
        errorMsg("Close failed: " + e.getMessage());
      }
    }
    ir = null;
    dir = null;
    is = null;
    removeAll();
    addComponent(this, "/xml/luke.xml", null, null);
    initPlugins();
  }
  
  public void actionReopen() {
    boolean mmap = dir instanceof MMapDirectory;
    openIndex(pName, false, mmap, readOnly);
  }
  /**
   * Open indicated index and re-initialize all GUI and plugins.
   * @param pName path to index
   * @param force if true, and the index is locked, unlock it first. If false, and
   * the index is locked, an error will be reported.
   * @param mmap use MMapDirectory instead of FSDirectory. This may fail if {@link SecurityManager}
   * is present.
   * @param readOnly open in read-only mode, and disallow modifications.
   */
  public void openIndex(String name, boolean force, boolean mmap, boolean ro) {
    pName = name;
    readOnly = ro;
    removeAll();
    this.baseDir = new File(name).getParent();
    addComponent(this, "/xml/luke.xml", null, null);
    statmsg = find("statmsg");
    if (dir != null) {
      try {
        if (ir != null) ir.close();
      } catch (Exception e) {}
      ;
      try {
        dir.close();
      } catch (Exception e) {}
      ;
    }
    try {
      dir = openDirectory(mmap, new File(pName), false);
      if (IndexReader.isLocked(dir)) {
        if (ro) {
          errorMsg("Index is locked and Read-Only. Open for read-write and 'Force unlock'.");
          dir.close();
          return;
        }
        if (force) {
          IndexReader.unlock(dir);
        } else {
          errorMsg("Index is locked. Try 'Force unlock' when opening.");
          dir.close();
          return;
        }
      }

      ir = IndexReader.open(dir);
      is = new IndexSearcher(dir);
      initOverview();
      initPlugins();
      showStatus("Index successfully open.");
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg(e.getMessage());
      return;
    }
  }

  /**
   * This method attempts to work around the limitation in FSDirectory which
   * prevents you from switching Directory implementation during lifetime of a
   * JVM. Unfortunately, this doesn't work if a SecurityManager is present.
   * 
   * @param mmap if true use MMapDirectory, otherwise use FSDirectory
   * @param file index directory
   * @param create if true, create a new directory
   * @return directory implementation
   */
  Class defaultDirImpl = null;
  
  public FSDirectory openDirectory(boolean mmap, File file, boolean create) throws Exception {
    FSDirectory res = null;
    java.lang.reflect.Field impl = FSDirectory.class.getDeclaredField("IMPL");
    impl.setAccessible(true);
    if (defaultDirImpl == null) {
      defaultDirImpl = (Class)impl.get(null);
    }
    if (mmap) {
      impl.set(null, MMapDirectory.class);
      /*
      // XXX this is silly. why can't I select different implementations
      // XXX during runtime???
      java.lang.reflect.Field dirs = FSDirectory.class.getDeclaredField("DIRECTORIES");
      dirs.setAccessible(true);
      Hashtable DIRECTORIES = (Hashtable) dirs.get(null);
      Method m_init = FSDirectory.class.getDeclaredMethod("init", new Class[] { File.class, Boolean.TYPE });
      Method m_create = FSDirectory.class.getDeclaredMethod("create", new Class[0]);
      m_init.setAccessible(true);
      m_create.setAccessible(true);
      java.lang.reflect.Field refCount = FSDirectory.class.getDeclaredField("refCount");
      refCount.setAccessible(true);
      // copied verbatim from FSDirectory.java:
      try {
        file = new File(file.getCanonicalPath());
        synchronized (DIRECTORIES) {
          res = (FSDirectory) DIRECTORIES.get(file);
          if (res == null) {
            try {
              res = (FSDirectory) MMapDirectory.class.newInstance();
            } catch (Exception e) {
              throw new RuntimeException("cannot load FSDirectory class: " + e.toString());
            }
            m_init.invoke(res, new Object[] { file, new Boolean(create) });
            DIRECTORIES.put(file, res);
          } else if (create) {
            m_create.invoke(res, new Object[0]);
          }
        }
        synchronized (res) {
          refCount.setInt(res, refCount.getInt(res) + 1);
        }
        return res;
      } catch (Exception e) {
        showStatus("MMapDirectory failed: " + e.getMessage());
        res = null;
      }*/
      res = FSDirectory.getDirectory(file);
      if (res != null) return res;
    } else {
      impl.set(null, defaultDirImpl);
    }
    // fall-back to FSDirectory.
    if (!mmap || res == null) return FSDirectory.getDirectory(file);
    return null;
  }
  
  /**
   * Initialize plugins. This method should always be called when a new index is open.
   *
   */
  public void initPlugins() {
    Object pluginsTabs = find("pluginsTabs");
    removeAll(pluginsTabs);
    for (int i = 0; i < plugins.size(); i++) {
      LukePlugin plugin = (LukePlugin) plugins.get(i);
      addPluginTab(pluginsTabs, plugin);
      plugin.setDirectory(dir);
      plugin.setIndexReader(ir);
      try {
        plugin.init();
      } catch (Exception e) {
        e.printStackTrace();
        showStatus("PLUGIN ERROR: " + e.getMessage());
      }
    }
  }

  /**
   * Initialize index overview and other GUI elements. This method is called always
   * when a new index is open.
   *
   */
  private void initOverview() {
    try {
      courier = new Font("Courier", getFont().getStyle(), getFont().getSize());
      setBoolean(find("bReload"), "enabled", true);
      setBoolean(find("bClose"), "enabled", true);
      Object cbType = find("cbType");
      populateAnalyzers(cbType);
      Object pOver = find("pOver");
      Object iName = find("idx");
      setString(iName, "text", pName + (readOnly ? " (R)" : ""));
      iName = find(pOver, "iName");
      setString(iName, "text", pName + (readOnly ? " (Read-Only)" : ""));
      Object iDocs = find(pOver, "iDocs");
      String numdocs = String.valueOf(ir.numDocs());
      setString(iDocs, "text", numdocs);
      iDocs = find("iDocs1");
      setString(iDocs, "text", String.valueOf(ir.maxDoc() - 1));
      Object dirImpl = find("dirImpl");
      setString(dirImpl, "text", dir.getClass().getName());
      Object iFields = find(pOver, "iFields");
      fn = ir.getFieldNames(IndexReader.FieldOption.ALL);
      if (fn.size() == 0) {
        errorMsg("Invalid or corrupted index.");
        return;
      }
      Object fList = find(pOver, "fList");
      Object defFld = find("defFld");
      Object fCombo = find("fCombo");
      java.util.Iterator it = fn.iterator();
      removeAll(fList);
      removeAll(fCombo);
      TreeSet fields = new TreeSet();
      while (it.hasNext()) {
        String fld = (String)it.next();
        fields.add(fld);
      }
      idxFields = (String[])fields.toArray(new String[0]);
      setString(iFields, "text", String.valueOf(idxFields.length));
      for (int k = 0; k < idxFields.length; k++) {
        Object item = create("item");
        add(fList, item);
        setString(item, "text", "<" + idxFields[k] + ">");
        putProperty(item, "fName", idxFields[k]);
        item = create("choice");
        add(defFld, item);
        setString(item, "text", idxFields[k]);
        item = create("choice");
        add(fCombo, item);
        setString(item, "text", idxFields[k]);
        putProperty(item, "fName", idxFields[k]);
      }
      setString(defFld, "text", idxFields[0]);
      Object iTerms = find(pOver, "iTerms");
      TermEnum te = ir.terms();
      numTerms = 0;
      while (te.next())
        numTerms++;
      te.close();
      setString(iTerms, "text", String.valueOf(numTerms));
      Object iMod = find(pOver, "iMod");
      setString(iMod, "text", new java.util.Date(IndexReader.lastModified(dir)).toString());
      Object iVer = find(pOver, "iVer");
      setString(iVer, "text", String.valueOf(ir.getCurrentVersion(dir)));
      Object iDel = find(pOver, "iDel");
      setString(iDel, "text", ir.hasDeletions() ? "Yes" : "No");
      setString(find("defFld"), "text", idxFields[0]);
      // Remove columns
      Object header = get(find("sTable"), "header");
      removeAll(header);
      Object c = create("column");
      setString(c, "text", "#");
      setInteger(c, "width", 40);
      add(header, c);
      c = create("column");
      setString(c, "text", "Score");
      setInteger(c, "width", 50);
      add(header, c);
      c = create("column");
      setString(c, "text", "Doc. Id");
      setInteger(c, "width", 60);
      add(header, c);
      for (int j = 0; j < idxFields.length; j++) {
        c = create("column");
        setString(c, "text", idxFields[j]);
        add(header, c);
      }
      actionTopTerms(find("nTerms"));
      showFiles(dir);
      Object fileSize = find("iFileSize");
      BigInteger totalFileSize = calcTotalFileSize(pName, dir);
      setString(fileSize, "text", normalizeSize(totalFileSize) + normalizeUnit(totalFileSize));
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg(e.getMessage());
    }
  }

  private String SEP_CHAR = System.getProperty("file.separator");

  private void showFiles(Directory dir) throws Exception {
    String[] files = dir.list();
    List segs = getIndexFileNames(dir);
    List dels = getIndexDeletableNames(dir);
    Object filesTable = find("filesTable");
    removeAll(filesTable);
    for (int i = 0; i < files.length; i++) {
      String filename;
      if (pName.endsWith(SEP_CHAR)) {
        filename = pName;
      } else {
        filename = pName + SEP_CHAR;
      }
      File file = new File(filename + files[i]);
      Object row = create("row");
      Object nameCell = create("cell");
      setString(nameCell, "text", files[i]);
      add(row, nameCell);
      Object sizeCell = create("cell");
      setString(sizeCell, "text", normalizeSize(file.length()));
      add(row, sizeCell);
      Object unitCell = create("cell");
      setString(unitCell, "text", normalizeUnit(file.length()));
      add(row, unitCell);
      boolean deletable = dels.contains(files[i].intern());
      boolean inuse = false;
      int idx = files[i].indexOf('.');
      if (idx != -1) {
        if (files[i].equals("segments.gen")) {
          inuse = true;
        } else {
          String prefix = files[i].substring(0, idx).intern();
          inuse = segs.contains(prefix);
        }
      } else if (files[i].startsWith("segments") || files[i].equals("deletable")) inuse = true;
      Object delCell = create("cell");
      setString(delCell, "text", deletable ? "YES" : "-");
      add(row, delCell);
      Object inuseCell = create("cell");
      setString(inuseCell, "text", inuse ? "YES" : "-");
      add(row, inuseCell);
      add(filesTable, row);
    }
  }

  private void syncMRU(Object path) {
    removeAll(path);
    for (Iterator iter = Prefs.getMruList().iterator(); iter.hasNext();) {
      String element = (String) iter.next();
      Object choice = create("choice");
      setString(choice, "text", element);
      add(path, choice);
    }
  }

  private String normalizeUnit(long len) {
    if (len == 1) {
      return " byte";
    } else if (len < 1024) {
      return " bytes";
    } else if (len < 51200000) {
      return " kB";
    } else {
      return " MB";
    }
  }

  private String normalizeUnit(BigInteger len) {
    if (len.compareTo(new BigInteger("1")) == 0) {
      return " byte";
    } else if (len.compareTo(new BigInteger("1024")) == -1) {
      return " bytes";
    } else if (len.compareTo(new BigInteger("51200000")) == -1) {
      return " kB";
    } else {
      return " MB";
    }
  }

  private String normalizeSize(BigInteger len) {
    if (len.compareTo(new BigInteger("1")) == 0) {
      return len.toString();
    } else if (len.compareTo(new BigInteger("1024")) == -1) {
      return len.toString();
    } else if (len.compareTo(new BigInteger("51200000")) == -1) {
      return len.divide(new BigInteger("1024")).toString();
    } else {
      return len.divide(new BigInteger("1048576")).toString();
    }
  }

  private BigInteger calcTotalFileSize(String path, Directory fsdir) {
    BigInteger totalFileSize = new BigInteger("0");
    String[] files = null;
    if (fsdir instanceof FSDirectory) {
      files = ((FSDirectory)fsdir).list();
    } else if (fsdir instanceof MMapDirectory) {
      files = ((MMapDirectory)fsdir).list();
    }
    if (files == null) return totalFileSize;
    for (int i = 0; i < files.length; i++) {
      String filename;
      if (pName.endsWith(SEP_CHAR)) {
        filename = pName;
      } else {
        filename = pName + SEP_CHAR;
      }

      File file = new File(filename + files[i]);
      totalFileSize = totalFileSize.add(new BigInteger(file.length() + ""));
    }
    return totalFileSize;
  }

  private String normalizeSize(long len) {
    if (len < 1024) {
      return len + "";
    } else if (len < 51200000) {
      return (len / 1024) + "";
    } else {
      return (len / 1048576) + "";
    }
  }

  /**
   * Update the list of top terms.
   * @param nTerms Thinlet widget containing the number of top terms to show
   */
  public void actionTopTerms(Object nTerms) {
    String sndoc = getString(nTerms, "text");
    int ndoc = 50;
    try {
      ndoc = Integer.parseInt(sndoc);
    } catch (Exception e) {}
    Object[] fields = getSelectedItems(find("fList"));
    String[] flds = null;
    if (fields == null || fields.length == 0) {
      flds = idxFields;
    } else {
      flds = new String[fields.length];
      for (int i = 0; i < fields.length; i++) {
        flds[i] = (String) getProperty(fields[i], "fName");
      }
    }
    try {
      TermInfo[] tis = HighFreqTerms.getHighFreqTerms(dir, null, ndoc + 1, flds);
      Object table = find("tTable");
      removeAll(table);
      if (tis == null || tis.length == 0) {
        Object row = create("row");
        Object cell = create("cell");
        add(row, cell);
        cell = create("cell");
        add(row, cell);
        cell = create("cell");
        add(row, cell);
        cell = create("cell");
        setBoolean(cell, "enabled", false);
        setString(cell, "text", "No Results");
        add(row, cell);
        add(table, row);
        return;
      }
      for (int i = 0; i < tis.length; i++) {
        Object row = create("row");
        add(table, row);
        putProperty(row, "term", tis[i].term);
        putProperty(row, "ti", tis[i]);
        Object cell = create("cell");
        setChoice(cell, "alignment", "right");
        setString(cell, "text", String.valueOf(i + 1));
        add(row, cell);
        cell = create("cell");
        setChoice(cell, "alignment", "right");
        setString(cell, "text", String.valueOf(tis[i].docFreq) + "  ");
        add(row, cell);
        cell = create("cell");
        setString(cell, "text", " <" + tis[i].term.field() + "> ");
        add(row, cell);
        cell = create("cell");
        setString(cell, "text", "  " + tis[i].term.text());
        add(row, cell);
      }
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg(e.getMessage());
    }
  }
  
  public void clipTopTerms(Object tTable) {
    Object[] rows = getItems(tTable);
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < rows.length; i++) {
      TermInfo ti = (TermInfo)getProperty(rows[i], "ti");
      if (ti == null) continue;
      sb.append(ti.docFreq + "\t" + ti.term.field() + "\t" + ti.term.text() + "\n");
    }
    StringSelection sel = new StringSelection(sb.toString());
    Toolkit.getDefaultToolkit().getSystemClipboard().setContents(sel, this);
  }

  /**
   * Switch to a view that shows all documents containing selected term.
   * @param tTable Thinlet table widget, where selected row contains a property named "term",
   * which is the selected {@link Term} instance
   */
  public void browseTermDocs(Object tTable) {
    Object row = getSelectedItem(tTable);
    if (row == null) return;
    Term t = (Term) getProperty(row, "term");
    if (t == null) return;
    Object tabpane = find("maintpane");
    setInteger(tabpane, "selected", 1);
    _showTerm(find("fCombo"), find("fText"), t);
    repaint();

  }

  public void showTermDocs(Object tTable) {
    Object row = getSelectedItem(tTable);
    if (row == null) return;
    Term t = (Term) getProperty(row, "term");
    if (t == null) return;
    Object tabpane = find("maintpane");
    setInteger(tabpane, "selected", 2);
    Object qField = find("qField");
    setString(qField, "text", t.field() + ":" + t.text());
    search(qField);
    repaint();

  }

  /**
   * Undelete all deleted documents in the current index. This method also
   * updates the overview.
   */
  public void actionUndelete() {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    if (readOnly) {
      showStatus(MSG_READONLY);
      return;
    }
    try {
      ir.undeleteAll();
      initOverview();
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg(e.getMessage());
    }
  }

  /** Not implemented yet... */
  public void actionConvert(Object method) {

  }
  
  private static Vector emptyDels = new Vector();

  public List getIndexDeletableNames(Directory d) {
    if (d == null) return null;
    Vector deletable = null;
    try {
      deletable = IndexGate.getDeletableFiles(d);
    } catch (Exception e) {
      e.printStackTrace();
    }
    //for (int i = 0; i < deletable.size(); i++) {
    //  System.out.println(" -del " + deletable.get(i));
    //}
    if (deletable == null) deletable = emptyDels;
    return deletable;
  }
  
  public List getIndexFileNames(Directory d) {
    if (d == null) return null;
    Vector names = null;
    try {
      names = IndexGate.getSegmentFiles(d);
    } catch (Exception e) {
      e.printStackTrace();
    }
    //for (int i = 0; i < names.size(); i++) {
    //  System.out.println(" -seg " + names.get(i));
    //}
    return names;
  }
  
  /**
   * This method will cleanup the current Directory of any content
   * that is not the part of the index. All files and directories,
   * which are not listed in the "segments" file, will be removed.
   * NOTE: all files marked as deletable in the "deletable" file will
   * also be removed, and the current Directory and IndexReader will be
   * reopened.
   *
   */
  public void actionCleanup() {
    if (dir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    try {
      ir.close();
    } catch (Exception e) {
      
    }
    // collect names belonging to this index
    List segs = getIndexFileNames(dir);
    List deletable = getIndexDeletableNames(dir);
    String clazz = dir.getClass().getName();
    boolean mmap = false;
    if (clazz.equals("org.apache.lucene.store.MMapDirectory")) mmap = true;
    Vector v = new Vector();
    Vector newDel = new Vector();
    try {
      File iDir = new File(pName);
      File[] files = iDir.listFiles();
      for (int i = 0; i < files.length; i++) {
        String name = files[i].getName();
        if (name.startsWith("segments") || name.equals("segments.gen") || name.equals("deletable")) continue;
        if (deletable.contains(name.intern())) {
          if (!removeFile(files[i])) {
            newDel.add(name);
          }
          continue;
        }
        int idx = name.indexOf('.');
        if (idx == -1) removeFile(files[i]);
        String prefix = name.substring(0, idx).intern();
        if (segs.contains(prefix)) continue;
        if (!removeFile(files[i])) v.add(files[i]);
      }
      // re-create deletable.
      if (newDel.size() > 0) {
        IndexOutput output = dir.createOutput("deletable.new");
        try {
          output.writeInt(newDel.size());
          for (int i = 0; i < newDel.size(); i++)
            output.writeString((String)newDel.elementAt(i));
        } finally {
          output.close();
        }
      }
      try {
        dir.close();
      } catch (Exception e) {
        e.printStackTrace();
      }
      File dold = new File(pName, "deletable");
      File dnew = new File(pName, "deletable.new");
      if (dold.exists()) dold.delete();
      if (dnew.exists()) dnew.renameTo(dold);
      dir = openDirectory(mmap, new File(pName), false);

      if (v.size() > 0) {
        StringBuffer msg = new StringBuffer("Some files could not be deleted:");
        for (int i = 0; i < v.size(); i++) msg.append("\n" + v.get(i));
        errorMsg(msg.toString());
      }
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg("FAILED: " + e.getMessage());
    } finally {
      try {
        ir = IndexReader.open(dir);
      } catch (Exception e) {
        e.printStackTrace();
        errorMsg(e.getMessage());
      }
      initOverview();
      initPlugins();
    }
  }
  
  /** Recursively remove files and directories including the indicated
   * root file.
   * @param f root file name
   * @return true if successful, false otherwise. Note that if the result
   * is false the tree may have been partially removed.
   */
  public boolean removeFile(File f) {
    System.out.println("remove " + f);
    if (!f.isDirectory()) return f.delete();
    File[] files = f.listFiles();
    boolean res = true;
    for (int i = 0; i < files.length; i++) {
      res = res && removeFile(files[i]);
    }
    return res;
  }
  
  /**
   * Optimize the current index
   * @param method Thinlet menuitem widget containing the choice of index format.
   * If the widget name is "optCompound" then the index will be optimized into compound
   * format; otherwise a plain multi-file format will be used.
   * <p>NOTE: this method is usually invoked from the GUI, and it also re-initializes GUI
   * and plugins.</p>
   */
  public void actionOptimize(Object method) {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    if (readOnly) {
      showStatus(MSG_READONLY);
      return;
    }
    try {
      String name = getString(method, "name");
      boolean useCompound;
      if (name.equals("optCompound"))
        useCompound = true;
      else useCompound = false;
      if (is != null) is.close();
      ir = optimize(ir, useCompound);
      is = new IndexSearcher(ir);
      initOverview();
      initPlugins();
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  /**
   * Optimize the index. NOTE: this method does NOT re-initialize GUI nor plugins.
   * @param reader index reader to optimize
   * @param compound if true, use compound file format
   * @return newly open index reader, which uses the same {@link Directory} instance as the
   * input reader.
   * @throws Exception
   */
  public IndexReader optimize(IndexReader reader, boolean compound) throws Exception {
    Directory d = reader.directory();
    reader.close();
    IndexWriter iw = null;
    try {
      iw = new IndexWriter(d, new org.apache.lucene.analysis.WhitespaceAnalyzer(), false);
      iw.setUseCompoundFile(compound);
      BigInteger startSize = calcTotalFileSize(pName, dir);
      long startTime = System.currentTimeMillis();
      iw.optimize();
      long endTime = System.currentTimeMillis();
      BigInteger endSize = calcTotalFileSize(pName, dir);
      BigInteger deltaSize = startSize.subtract(endSize);
      String sign = deltaSize.compareTo(new BigInteger("0")) < 0 ? " Increased " : " Reduced ";
      String sizeMsg = sign + normalizeSize(deltaSize.abs()) + normalizeUnit(deltaSize.abs());
      String timeMsg = String.valueOf(endTime - startTime) + " ms";
      showStatus(sizeMsg + " in " + timeMsg);
      iw.close();
    } catch (Exception e) {
      if (iw != null) try {
        iw.close();
      } catch (Exception e1) {}
    }
    reader = IndexReader.open(d);
    return reader;
  }

  public void showPrevDoc(Object docNum) {
    _showDoc(docNum, -1);
  }

  public void showNextDoc(Object docNum) {
    _showDoc(docNum, +1);
  }

  public void showDoc(Object docNum) {
    _showDoc(docNum, 0);
  }

  private void _showDoc(Object docNum, int incr) {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    String num = getString(docNum, "text");
    if (num.trim().equals("")) num = String.valueOf(-incr);
    try {
      int iNum = Integer.parseInt(num);
      iNum += incr;
      if (iNum < 0 || iNum >= ir.maxDoc()) {
        showStatus("Document number outside valid range.");
        return;
      }
      setString(docNum, "text", String.valueOf(iNum));
      Document doc = null;
      if (!ir.isDeleted(iNum)) {
        doc = ir.document(iNum);
        _showDocFields(iNum, doc);
      } else {
        showStatus("Deleted document - not available.");
        _showDocFields(iNum, null);
      }
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }
  
  public void actionAddDocument(Object docTable) {
    Object dialog = addComponent(null, "/xml/editdoc.xml", null, null);
    remove(find(dialog, "bDelAdd"));
    Object cbAnalyzers = find(dialog, "cbAnalyzers");
    populateAnalyzers(cbAnalyzers);
    setInteger(cbAnalyzers, "selected", 0);
    add(dialog);
  }

  public void actionReconstruct(Object docTable) {
    final Integer DocNum = (Integer) getProperty(docTable, "docNum");
    final Document document = (Document) getProperty(docTable, "doc");
    if (DocNum == null || document == null) {
      showStatus("FAILED: selected doc. not available!");
      return;
    }
    final Object progress = addComponent(this, "/xml/progress.xml", null, null);
    setString(find(progress, "msg"), "text", "Collecting terms...");
    setInteger(find(progress, "bar"), "maximum", 100);
    Thread thr = new Thread() {
      public void run() {
        int docNum = DocNum.intValue();
        TreeMap doc = new TreeMap();
        // get stored fields
        Vector sf = new Vector();
        for (int i = 0; i < idxFields.length; i++) {
          Field[] f = document.getFields(idxFields[i]);
          if (f == null || !f[0].isStored()) continue;
          StringBuffer sb = new StringBuffer();
          for (int k = 0; k < f.length; k++) {
            if (k > 0) sb.append('\n');
            sb.append(f[k].stringValue());
          }
          Field field = new Field(idxFields[i], sb.toString(),
              f[0].isStored() ? Field.Store.YES : Field.Store.NO,
                  f[0].isIndexed() == false ? Field.Index.NO : (f[0].isTokenized() ? Field.Index.TOKENIZED : Field.Index.UN_TOKENIZED),
                  f[0].isTermVectorStored() ? Field.TermVector.YES : Field.TermVector.NO);
          field.setBoost(f[0].getBoost());
          doc.put("sf_" + idxFields[i], field);
          sf.add(idxFields[i]);
        }
        String term = null;
        GrowableStringArray terms = null;
        try {
          int i = 0;
          int delta = numTerms / 100;
          if (delta == 0) delta = 1;
          Object bar = find(progress, "bar");
          TermEnum te = ir.terms();
          TermPositions tp = ir.termPositions();
          while (te.next()) {
            if ((i++ % delta) == 0) {
              setInteger(bar, "value", i / delta);
            }
            // skip stored fields
            //if (sf.contains(te.term().field())) continue;
            tp.seek(te.term());
            if (!tp.skipTo(docNum) || tp.doc() != docNum) {
              // this term is not found in the doc
              continue;
            }
            term = te.term().text();
            terms = (GrowableStringArray) doc.get("uf_" + te.term().field());
            if (terms == null) {
              terms = new GrowableStringArray();
              doc.put("uf_" + te.term().field(), terms);
            }
            for (int k = 0; k < tp.freq(); k++) {
              int pos = tp.nextPosition();
              terms.append(pos, "|", term);
            }
          }
          Object dialog = addComponent(null, "/xml/editdoc.xml", null, null);
          putProperty(dialog, "docNum", new Integer(docNum));
          Object cbAnalyzers = find(dialog, "cbAnalyzers");
          populateAnalyzers(cbAnalyzers);
          setInteger(cbAnalyzers, "selected", 0);
          Object editTabs = find(dialog, "editTabs");
          setString(find(dialog, "docNum"), "text", "Fields of Doc #: " + docNum);
          for (int p = 0; p < idxFields.length; p++) {
            String key = idxFields[p];
            if (doc.get("sf_" + key) == null && doc.get("uf_" + key) == null) continue;
            Object tab = create("tab");
            setString(tab, "text", key);
            setFont(tab, getFont().deriveFont(Font.BOLD));
            add(editTabs, tab);
            Object editfield = addComponent(tab, "/xml/editfield.xml", null, null);
            Object t = doc.get("sf_" + key);
            Object ut = doc.get("uf_" + key);
            Object fType = find(editfield, "fType");
            Object sText = find(editfield, "sText");
            Object rText = find(editfield, "rText");
            Object fBoost = find(editfield, "fBoost");
            Object cbStored = find(editfield, "cbStored");
            Object cbIndexed = find(editfield, "cbIndexed");
            Object cbTokenized = find(editfield, "cbTokenized");
            Object cbTVF = find(editfield, "cbTVF");
            Object stored = find(editfield, "stored");
            Object restored = find(editfield, "restored");
            if (t != null) {
              Field f = (Field) t;
              setString(fType, "text", "Original stored field content");
              setString(sText, "text", f.stringValue());
              setString(fBoost, "text", String.valueOf(f.getBoost()));
              setBoolean(cbStored, "selected", f.isStored());
              setBoolean(cbIndexed, "selected", f.isIndexed());
              setBoolean(cbTokenized, "selected", f.isTokenized());
              setBoolean(cbTVF, "selected", f.isTermVectorStored());
            } else {
              remove(stored);
            }
            if (ut != null) {
              String sep = " ";
              if (t == null) {
                setString(fType, "text", "RESTORED content ONLY - check for errors!");
                setColor(fType, "foreground", Color.red);
              } else {
                setBoolean(rText, "editable", false);
                setBoolean(rText, "border", false);
                setString(restored, "text", "Tokenized");
                sep = ", ";
              }
              setBoolean(cbIndexed, "selected", true);
              setString(fBoost, "text", String.valueOf(document.getBoost()));
              terms = (GrowableStringArray) doc.get("uf_" + key);
              if (terms == null) continue;
              StringBuffer sb = new StringBuffer();
              String sNull = "null";
              int k = 0, m = 0;
              for (int j = 0; j < terms.size(); j++) {
                if (terms.get(j) == null)
                  k++;
                else {
                  if (sb.length() > 0) sb.append(sep);
                  if (m > 0 && m % 5 == 0) sb.append('\n');
                  if (k > 0) {
                    sb.append(sNull + "_" + k + sep);
                    k = 0;
                    m++;
                  }
                  sb.append(escape(terms.get(j)));
                  //else sb.append(terms.get(j));
                  m++;
                }
              }
              setString(rText, "text", sb.toString());
            } else {
              remove(restored);
            }
          }
          add(dialog);
          getPreferredSize(editTabs);
        } catch (Exception e) {
          e.printStackTrace();
          showStatus(e.getMessage());
        }
        remove(progress);
      }
    };
    thr.start();
  }

  public boolean actionEditAdd(Object editdoc) {
    Document doc = new Document();
    Object cbAnalyzers = find(editdoc, "cbAnalyzers");
    Analyzer a = new StandardAnalyzer();
    String clazz = null;
    try {
      clazz = getString(cbAnalyzers, "text");
      a = (Analyzer) Class.forName(clazz).newInstance();
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg("Could not instantiate Analyzer '" + clazz + "'. Please select a valid Analyzer.");
      return false;
    }
    Object editTabs = find(editdoc, "editTabs");
    Object[] tabs = getItems(editTabs);
    for (int i = 0; i < tabs.length; i++) {
      String name = getString(tabs[i], "text");
      if (name.trim().equals("")) continue;
      Object fBoost = find(tabs[i], "fBoost");
      Object fText = find(tabs[i], "sText");
      if (fText == null) fText = find(tabs[i], "rText");
      Object cbStored = find(tabs[i], "cbStored");
      Object cbIndexed = find(tabs[i], "cbIndexed");
      Object cbTokenized = find(tabs[i], "cbTokenized");
      Object cbTVF = find(tabs[i], "cbTVF");
      Object cbCmp = find(tabs[i], "cbCmp");
      Object cbONorms = find(tabs[i], "cbONorms");
      String text = getString(fText, "text");
      Field f = new Field(name, text,
          getBoolean(cbStored, "selected") ? (getBoolean(cbCmp, "selected") ? Field.Store.COMPRESS : Field.Store.YES) : Field.Store.NO, 
              getBoolean(cbIndexed, "selected") ?
                  (getBoolean(cbTokenized, "selected") ? Field.Index.TOKENIZED : Field.Index.UN_TOKENIZED) : Field.Index.NO,
                  getBoolean(cbTVF, "selected") ? Field.TermVector.YES : Field.TermVector.NO);
      f.setOmitNorms(getBoolean(cbONorms, "selected"));
      String boostS = getString(fBoost, "text").trim();
      if (!boostS.equals("") && !boostS.equals("1.0")) {
        float boost = 1.0f;
        try {
          boost = Float.parseFloat(boostS);
        } catch (Exception e) {
          e.printStackTrace();
        }
        f.setBoost(boost);
      }
      doc.add(f);
    }
    IndexWriter writer = null;
    boolean res = false;
    String msg = null;
    try {
      ir.close();
      writer = new IndexWriter(dir, a, false);
      writer.addDocument(doc);
      res = true;
    } catch (Exception e) {
      e.printStackTrace();
      msg = "FAILED: " + e.getMessage();
      res = false;
    } finally {
      try {
        if (writer != null) writer.close();
      } catch (Exception e) {
        e.printStackTrace();
        res = false;
        msg = "FAILED: " + e.getMessage();
      }
      try {
        ir = IndexReader.open(dir);
      } catch (Exception e) {
        e.printStackTrace();
        res = false;
        msg = "FAILED: " + e.getMessage();
      }
      remove(editdoc);
      initOverview();
      initPlugins();
    }
    if (!res) {
      errorMsg(msg);
    }
    return res;
  }

  public void actionEditReplace(Object editdoc) {
    if (!actionEditAdd(editdoc)) return;
    Integer DocNum = (Integer) getProperty(editdoc, "docNum");
    if (DocNum == null) return;
    try {
      ir.deleteDocument(DocNum.intValue());
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg("ERROR deleting: " + e.getMessage());
      return;
    }
  }

  public void actionEditAddField(Object editdoc) {
    String name = getString(find(editdoc, "fNewName"), "text");
    if (name.trim().equals("")) {
      showStatus("FAILED: Field name is required.");
      return;
    }
    name = name.trim();
    Object editTabs = find(editdoc, "editTabs");
    Object tab = create("tab");
    setString(tab, "text", name);
    setFont(tab, getFont().deriveFont(Font.BOLD));
    add(editTabs, tab);
    Object editfield = addComponent(tab, "/xml/editfield.xml", null, null);
    repaint(editTabs);
  }

  public void actionEditDeleteField(Object editfield) {
    Object tab = getParent(editfield);
    remove(tab);
  }

  private void _showDocFields(int docid, Document doc) {
    Object table = find("docTable");
    setString(find("docNum"), "text", String.valueOf(docid));
    removeAll(table);
    putProperty(table, "doc", doc);
    putProperty(table, "docNum", new Integer(docid));
    if (doc == null) {
      setString(find("docNum1"), "text", String.valueOf(docid) + " UNAVAILABLE");
      setString(find("docBoost"), "text", "?");
      return;
    }
    setString(find("docNum1"), "text", String.valueOf(docid));
    setString(find("docBoost"), "text", String.valueOf(doc.getBoost()));
    for (int i = 0; i < idxFields.length; i++) {
      Field[] fields = doc.getFields(idxFields[i]);
      if (fields == null) {
        addFieldRow(table, idxFields[i], null);
        continue;
      }
      for (int j = 0; j < fields.length; j++) {
        addFieldRow(table, idxFields[i], fields[j]);
      }
    }
    doLayout(table);
  }

  Font courier = null;
  private void addFieldRow(Object table, String fName, Field f) {
    Object row = create("row");
    add(table, row);
    putProperty(row, "field", f);
    putProperty(row, "fName", fName);
    Object cell = create("cell");
    setString(cell, "text", "<" + fName + ">");
    add(row, cell);
    cell = create("cell");
    StringBuffer flags = new StringBuffer();
    if (f != null && f.isIndexed()) flags.append("I");
    else flags.append("-");
    if (f != null && f.isTokenized()) flags.append("T");
    else flags.append("-");
    if (f != null && f.isStored()) flags.append("S");
    else flags.append("-");
    if (f != null && f.isTermVectorStored()) flags.append("V");
    else flags.append("-");
    if (f != null && f.isStoreOffsetWithTermVector()) flags.append("o");
    else flags.append("-");
    if (f != null && f.isStorePositionWithTermVector()) flags.append("p");
    else flags.append("-");
    if (f != null && f.getOmitNorms()) flags.append("O");
    else flags.append("-");
    if (f != null && f.isLazy()) flags.append("L");
    else flags.append("-");
    if (f != null && f.isBinary()) flags.append("B");
    else flags.append("-");
    if (f != null && f.isCompressed()) flags.append("C");
    else flags.append("-");
    setString(cell, "text", flags.toString());
    setFont(cell, "font", courier);
    add(row, cell);
    cell = create("cell");
    if (f != null) setString(cell, "text", String.valueOf(f.getBoost()));
    add(row, cell);
    cell = create("cell");
    if (f != null)
      setString(cell, "text", escape(f.stringValue()));
    else {
      setString(cell, "text", "<not available>");
      setBoolean(cell, "enabled", false);
    }
    add(row, cell);
  }

  public void showTV(Object table) {
    Object row = getSelectedItem(table);
    if (row == null) return;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    Integer DocId = (Integer) getProperty(table, "docNum");
    if (DocId == null) {
      showStatus("Missing Doc. Id.");
      return;
    }
    try {
      String fName = (String) getProperty(row, "fName");
      TermFreqVector tfv = ir.getTermFreqVector(DocId.intValue(), fName);
      if (tfv == null) {
        showStatus("Term Vector not available.");
        return;
      }
      Object dialog = addComponent(null, "/xml/vector.xml", null, null);
      setString(find(dialog, "fld"), "text", fName);
      Object vTable = find(dialog, "vTable");
      IntPair[] tvs = new IntPair[tfv.size()];
      String[] terms = tfv.getTerms();
      int[] freqs = tfv.getTermFrequencies();
      for (int i = 0; i < terms.length; i++) {
        IntPair ip = new IntPair(freqs[i], terms[i]);
        tvs[i] = ip;
      }
      Arrays.sort(tvs, new IntPair.PairComparator(false, true));
      for (int i = 0; i < tvs.length; i++) {
        Object r = create("row");
        add(vTable, r);
        putProperty(r, "tf", tvs[i]);
        Object cell = create("cell");
        setString(cell, "text", String.valueOf(tvs[i].cnt));
        add(r, cell);
        cell = create("cell");
        setString(cell, "text", escape(tvs[i].text));
        add(r, cell);
      }
      add(dialog);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }
  
  public void clipTV(Object vTable) {
    Object[] rows = getItems(vTable);
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < rows.length; i++) {
      IntPair tf = (IntPair)getProperty(rows[i], "tf");
      sb.append(tf.cnt + "\t" + tf.text + "\n");
    }
    StringSelection sel = new StringSelection(sb.toString());
    Toolkit.getDefaultToolkit().getSystemClipboard().setContents(sel, this);
  }

  public void clipCopyFields(Object table) {
    Object[] rows = getSelectedItems(table);
    if (rows == null || rows.length == 0) return;
    Document doc = (Document) getProperty(table, "doc");
    if (doc == null) return;
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < rows.length; i++) {
      Field f = (Field) getProperty(rows[i], "field");
      if (f == null) continue;
      if (i > 0) sb.append('\n');
      sb.append(f.toString());
    }
    StringSelection sel = new StringSelection(sb.toString());
    Toolkit.getDefaultToolkit().getSystemClipboard().setContents(sel, this);
  }

  public void clipCopyDoc(Object table) {
    Document doc = (Document) getProperty(table, "doc");
    if (doc == null) return;
    StringBuffer sb = new StringBuffer();
    Object[] rows = getItems(table);
    for (int i = 0; i < rows.length; i++) {
      Field f = (Field) getProperty(rows[i], "field");
      if (f == null) continue;
      if (i > 0) sb.append('\n');
      sb.append(f.toString());
    }
    StringSelection sel = new StringSelection(sb.toString());
    Toolkit.getDefaultToolkit().getSystemClipboard().setContents(sel, this);
  }

  public void showFirstTerm(Object fCombo, Object fText) {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    try {
      String fld = getString(fCombo, "text");
      TermEnum te = ir.terms(new Term(fld, ""));
      Term t = te.term();
      _showTerm(fCombo, fText, t);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  public void showNextTerm(Object fCombo, Object fText) {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    try {
      String text = getString(fText, "text");
      String fld = getString(fCombo, "text");
      TermEnum te = null;
      if (text == null || text.trim().equals("")) text = "";
      te = ir.terms(new Term(fld, text));
      te.next();
      Term t = te.term();
      _showTerm(fCombo, fText, t);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  public void showTerm(Object fCombo, Object fText) {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    try {
      String text = getString(fText, "text");
      String fld = getString(fCombo, "text");
      if (text == null || text.trim().equals("")) return;
      Term t = new Term(fld, text);
      _showTerm(fCombo, fText, t);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  private void _showTerm(Object fCombo, Object fText, Term t) {
    if (t == null) {
      showStatus("No terms?!");
      return;
    }
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    Object[] choices = getItems(fCombo);
    for (int i = 0; i < choices.length; i++) {
      if (t.field().equals(getString(choices[i], "text"))) {
        setInteger(fCombo, "selected", i);
        break;
      }
    }
    setString(fText, "text", t.text());
    putProperty(fText, "term", t);
    putProperty(fText, "td", null);
    setString(find("tdNum"), "text", "?");
    setString(find("tFreq"), "text", "?");
    Object dFreq = find("dFreq");
    try {
      int freq = ir.docFreq(t);
      setString(dFreq, "text", String.valueOf(freq));
      dFreq = find("tdMax");
      setString(dFreq, "text", String.valueOf(freq));
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
      setString(dFreq, "text", "?");
    }
  }

  public void showFirstTermDoc(Object fText) {
    Term t = (Term) getProperty(fText, "term");
    if (t == null) return;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    try {
      TermDocs td = ir.termDocs(t);
      td.next();
      setString(find("tdNum"), "text", "1");
      putProperty(fText, "td", td);
      _showTermDoc(fText, td);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  public void showNextTermDoc(Object fText) {
    Term t = (Term) getProperty(fText, "term");
    if (t == null) return;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    try {
      TermDocs td = (TermDocs) getProperty(fText, "td");
      if (td == null) {
        showFirstTermDoc(fText);
        return;
      }
      if (!td.next()) return;
      Object tdNum = find("tdNum");
      String sCnt = getString(tdNum, "text");
      int cnt = 1;
      try {
        cnt = Integer.parseInt(sCnt);
      } catch (Exception e) {}
      ;
      setString(tdNum, "text", String.valueOf(cnt + 1));
      _showTermDoc(fText, td);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  public void showAllTermDoc(Object fText) {
    Term t = (Term) getProperty(fText, "term");
    if (t == null) return;
    if (ir == null) {
      showStatus("MSG_NOINDEX");
      return;
    }
    Object tabpane = find("maintpane");
    setInteger(tabpane, "selected", 2);
    Object qField = find("qField");
    setString(qField, "text", t.field() + ":" + t.text());
    Object qFieldParsed = find("qFieldParsed");
    Query q = new TermQuery(t);
    setString(qFieldParsed, "text", q.toString());
    IndexSearcher is = null;
    try {
      is = new IndexSearcher(dir);
      Object sTable = find("sTable");
      removeAll(sTable);
      AllHitsCollector ahc = new AllHitsCollector();
      _search(q, is, ahc, sTable);
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg(e.getMessage());
    } finally {
      if (is != null) try {
        is.close();
      } catch (Exception e1) {}
      ;
    }
  }

  /**
   * Create a QueryParser instance that corresponds to values selected in the UI,
   * such as analyzer class name and arguments, and default field.
   * @return
   */
  public QueryParser createQueryParser() {
    String sAnal = getString(find("cbType"), "text");
    if (sAnal.trim().equals("")) {
      sAnal = "org.apache.lucene.analysis.standard.StandardAnalyzer";
      setString(find("cbType"), "text", sAnal);
    }
    String arg = null;
    if (sAnal.equals("org.apache.lucene.analysis.snowball.SnowballAnalyzer")) {
      arg = getString(find("snoName"), "text");
    }
    try {
      if (arg == null) {
        analyzer = (Analyzer) Class.forName(sAnal)
            .getConstructor(new Class[0])
            .newInstance(new Object[0]);
      } else {
        analyzer = (Analyzer) Class.forName(sAnal)
            .getConstructor(new Class[]{String.class})
            .newInstance(new Object[]{arg});
      }
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg("Analyzer '" + sAnal + "' error: " + e.getMessage() + ". Using StandardAnalyzer.");
      analyzer = stdAnalyzer;
    }
    String defField = getString(find("defFld"), "text");
    if (defField == null || defField.trim().equals("")) {
      defField = idxFields[0];
      setString(find("defFld"), "text", defField);
    }
    return new QueryParser(defField, analyzer);
  }

  public void explainStructure(Object qTabs) {
    showParsed();
    int idx = getSelectedIndex(qTabs);
    Object qField = find("qField");
    Query q = null;
    if (idx == 0) {
      q = (Query)getProperty(qField, "qParsed");
    } else {
      q = (Query)getProperty(qField, "qRewritten");
    }
    Object dialog = addComponent(this, "/xml/qexplain.xml", null, null);
    Object tree = find(dialog, "qTree");
    _explainStructure(tree, q);
  }
  
  private void _explainStructure(Object node, Query q) {
    String clazz = q.getClass().getName();
    int idx = clazz.lastIndexOf('.');
    clazz = clazz.substring(idx + 1);
    float boost = q.getBoost();
    Object n = create("node");
    add(node, n);
    String msg = clazz + ": boost=" + df.format(boost);
    setFont(n, getFont().deriveFont(Font.BOLD));
    setString(n, "text", msg);
    if (clazz.equals("TermQuery")) {
      Object n1 = create("node");
      Term t = ((TermQuery)q).getTerm();
      setString(n1, "text", "Term: field='" + t.field() + "' text='" + t.text() + "'");
      add(n, n1);
    } else if (clazz.equals("BooleanQuery")) {
      BooleanQuery bq = (BooleanQuery)q;
      BooleanClause[] clauses = bq.getClauses();
      int max = bq.getMaxClauseCount();
      boolean sc14 = bq.getUseScorer14();
      Object n1 = create("node");
      setString(n1, "text", "clauses=" + clauses.length +
              ", maxClauses=" + max +
              ", scorer14=" + sc14);
      add(n, n1);
      for (int i = 0; i < clauses.length; i++) {
        n1 = create("node");
        setString(n1, "text", "Clause " + i + ": " + clauses[i].getOccur().toString());
        add(n, n1);
        _explainStructure(n1, clauses[i].getQuery());
      }
    } else if (clazz.equals("PrefixQuery")) {
      Object n1 = create("node");
      Term t = ((PrefixQuery)q).getPrefix();
      setString(n1, "text", "Prefix: field='" + t.field() + "' text='" + t.text() + "'");
      add(n, n1);
    } else if (clazz.equals("PhraseQuery")) {
      PhraseQuery pq = (PhraseQuery)q;
      setString(n, "text", getString(n, "text") + ", slop=" + pq.getSlop());
      int[] pos = pq.getPositions();
      Term[] terms = pq.getTerms();
      Object n1 = create("node");
      StringBuffer sb = new StringBuffer("pos: [");
      for (int i = 0; i < pos.length; i++) {
        if (i > 0) sb.append(',');
        sb.append("" + pos[i]);
      }
      sb.append("]");
      setString(n1, "text", sb.toString());
      add(n, n1);
      for (int i = 0; i < terms.length; i++) {
        n1 = create("node");
        setString(n1, "text", "Term " + i + ": field='" + terms[i].field() +
                "' text='" + terms[i].text() + "'");
        add(n, n1);
      }
    } else if (clazz.equals("MultiPhraseQuery")) {
      MultiPhraseQuery pq = (MultiPhraseQuery)q;
      setString(n, "text", getString(n, "text") + ", slop=" + pq.getSlop());
      int[] pos = pq.getPositions();
      Object n1 = create("node");
      StringBuffer sb = new StringBuffer("pos: [");
      for (int i = 0; i < pos.length; i++) {
        if (i > 0) sb.append(',');
        sb.append("" + pos[i]);
      }
      sb.append("]");
      setString(n1, "text", sb.toString());
      add(n, n1);
      n1 = create("node");
      System.err.println("MultiPhraseQuery is missing the public getTermArrays() :-(");
      setString(n1, "text", "toString: " + pq.toString());
      add(n, n1);
    } else if (clazz.equals("FuzzyQuery")) {
      FuzzyQuery fq = (FuzzyQuery)q;
      Object n1 = create("node");
      setString(n1, "text", "prefixLen=" + fq.getPrefixLength() +
              ", minSimilarity=" + df.format(fq.getMinSimilarity()));
      add(n, n1);
      // do some tricks with reflection...
      try {
        Method m = FuzzyQuery.class.getDeclaredMethod("getEnum", new Class[]{IndexReader.class});
        m.setAccessible(true);
        FilteredTermEnum fte = (FilteredTermEnum)m.invoke(fq, new Object[]{ir});
        n1 = create("node");
        String clz = fte.getClass().getName();
        setString(n1, "text", clz + ": diff=" + df.format(fte.difference()));
        add(n, n1);
        do {
          n1 = create("node");
          Term t = fte.term();
          setString(n1, "text", "Term: field='" + t.field() +
                  "' text='" + t.text() + "', docFreq=" + fte.docFreq());
          add(n, n1);
        } while (fte.next());
      } catch (Exception e) {
        n1 = create("node");
        setString(n1, "text", "FilteredTermEnum: Exception " + e.getMessage());
        add(n, n1);
      }
    } else if (clazz.equals("WildcardQuery")) {
      WildcardQuery wq = (WildcardQuery)q;
      Term t = wq.getTerm();
      setString(n, "text", getString(n, "text") + ", term=" + t);
      // do some tricks with reflection...
      try {
        Method m = WildcardQuery.class.getDeclaredMethod("getEnum", new Class[]{IndexReader.class});
        m.setAccessible(true);
        FilteredTermEnum fte = (FilteredTermEnum)m.invoke(wq, new Object[]{ir});
        Object n1 = create("node");
        String clz = fte.getClass().getName();
        setString(n1, "text", clz + ": diff=" + df.format(fte.difference()));
        add(n, n1);
        do {
          n1 = create("node");
          t = fte.term();
          if (t == null) continue;
          setString(n1, "text", "Term: field='" + t.field() +
                  "' text='" + t.text() + "', docFreq=" + fte.docFreq());
          add(n, n1);
        } while (fte.next());
      } catch (Exception e) {
        Object n1 = create("node");
        setString(n1, "text", "FilteredTermEnum: Exception " + e.getMessage());
        add(n, n1);
      }
    } else if (clazz.equals("RangeQuery")) {
      RangeQuery rq = (RangeQuery)q;
      setString(n, "text", getString(n, "text") + ", inclusive=" + rq.isInclusive());
      Object n1 = create("node");
      setString(n1, "text", "lowerTerm: field='" + rq.getLowerTerm().field() +
              "' text='" + rq.getLowerTerm().text() + "'");
      add(n, n1);
      n1 = create("node");
      setString(n1, "text", "upperTerm: field='" + rq.getUpperTerm().field() +
              "' text='" + rq.getUpperTerm().text() + "'");
      add(n, n1);
    } else if (q instanceof SpanQuery) {
      SpanQuery sq = (SpanQuery)q;
      setString(n, "text", getString(n, "text") + ", field=" + sq.getField());
      Collection terms = sq.getTerms();
      Object n1 = null;
      if (terms != null) {
        n1 = create("node");
        setString(n1, "text", "Matched terms (" + terms.size() + "):");
        add(n, n1);
        Iterator it = terms.iterator();
        while(it.hasNext()) {
          Object n2 = create("node");
          Term t = (Term)it.next();
          setString(n2, "text", "field='" + t.field() + "' text='" + t.text() + "'");
          add(n1, n2);
        }
      } else {
        n1 = create("node");
        setString(n1, "text", "<no terms matched>");
        add(n, n1);
      }
      try {
        Spans spans = sq.getSpans(ir);
        n1 = create("node");
        setString(n1, "text", "Spans:");
        add(n, n1);
        do {
          Object n2 = create("node");
          setString(n2, "text", "doc=" + spans.doc() +
                  ", start=" + spans.start() + ", end=" + spans.end());
          add(n1, n2);
        } while (spans.next());
      } catch (Exception e) {
        n1 = create("node");
        setString(n1, "text", "Spans Exception: " + e.getMessage());
        add(n, n1);
      }
    }
  }
  /**
   * Update the parsed and rewritten query views.
   *
   */
  public void showParsed() {
    QueryParser qp = createQueryParser();
    Object qField = find("qField");
    Object qFieldParsed = find("qFieldParsed");
    Object qFieldRewritten = find("qFieldRewritten");
    String queryS = getString(qField, "text");
    if (queryS.trim().equals("")) {
      setString(qFieldParsed, "text", "<empty query>");
      setBoolean(qFieldParsed, "enabled", false);
      return;
    } else {
      setBoolean(qFieldParsed, "enabled", true);
    }
    try {
      Query q = qp.parse(queryS);
      setString(qFieldParsed, "text", q.toString());
      putProperty(qField, "qParsed", q);
      q = q.rewrite(ir);
      setString(qFieldRewritten, "text", q.toString());
      putProperty(qField, "qRewritten", q);
    } catch (Throwable t) {
      setString(qFieldParsed, "text", t.getMessage());
    }
  }

  private static class AllHitsCollector extends HitCollector {
    private ArrayList hits = new ArrayList();
    
    public AllHitsCollector() {
    }
    
    public void collect(int doc, float score) {
      hits.add(new AllHit(doc, score));
    }
    
    public int getSize() {
      return hits.size();
    }
    
    public int getDocId(int i) {
      return ((AllHit)hits.get(i)).docId;
    }

    public float getScore(int i) {
      return ((AllHit)hits.get(i)).score;
    }

    private static class AllHit {
      public int docId;
      public float score;
      
      public AllHit(int docId, float score) {
        this.docId = docId;
        this.score = score;
      }
    }
  }
  
  /**
   * Perform a search. NOTE: this method is usually invoked from the GUI.
   * @param qField Thinlet widget containing the query
   */
  public void search(Object qField) {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    String queryS = getString(qField, "text");
    if (queryS.trim().equals("")) {
      showStatus("FAILED: Empty query.");
      return;
    }
    qp = createQueryParser();
    Object cbSimCust = find("cbSimCust");
    boolean useCustom = getBoolean(cbSimCust, "selected") && similarity != null;
    boolean allRes = getBoolean(find("cbAllRes"), "selected");
    Object sTable = find("sTable");
    removeAll(sTable);
    try {
      Query q = qp.parse(queryS);
      if (useCustom) is.setSimilarity(similarity);
      showParsed();
      AllHitsCollector ahc = null;
      if (allRes) {
        ahc = new AllHitsCollector();
      }
      _search(q, is, ahc, sTable);
    } catch (Throwable e) {
      e.printStackTrace();
      errorMsg(e.getMessage());
    }
  }
  
  int resStart = 0;
  int resCount = 20;

  private void _search(Query q, IndexSearcher is, AllHitsCollector ahc, Object sTable) throws Exception {
    long startTime = System.currentTimeMillis();
    Hits hits = null;
    if (ahc == null) hits = is.search(q);
    else is.search(q, ahc);
    long endTime = System.currentTimeMillis();
    showStatus(endTime - startTime + " ms");
    Object bsPrev = find("bsPrev");
    Object bsNext = find("bsNext");
    setBoolean(bsNext, "enabled", false);
    setBoolean(bsPrev, "enabled", false);
    boolean noResults = (ahc == null) ? (hits == null || hits.length() == 0) : (ahc.getSize() == 0);
    if (noResults) {
      Object row = create("row");
      Object cell = create("cell");
      add(sTable, row);
      add(row, cell);
      cell = create("cell");
      add(row, cell);
      cell = create("cell");
      setString(cell, "text", "No Results");
      setBoolean(cell, "enabled", false);
      add(row, cell);
      setString(find("resNum"), "text", "0");
      return;
    }

    int resNum = (ahc == null) ? hits.length() : ahc.getSize();
    if (resNum > resCount) {
      setBoolean(bsNext, "enabled", true);
    }
    setString(find("resNum"), "text", String.valueOf(resNum));
    putProperty(sTable, "resNum", new Integer(resNum));
    putProperty(sTable, "query", q);
    putProperty(sTable, "hits", hits);
    putProperty(sTable, "ahc", ahc);
    resStart = 0;
    _showSearchPage(sTable);
  }
  
  public void prevPage(Object sTable) {
    int resNum = ((Integer)getProperty(sTable, "resNum")).intValue();
    if (resStart == 0) {
      setBoolean(find("bsPrev"), "enabled", false);
      return;
    }
    resStart -= resCount;
    if (resStart < 0) resStart = 0;
    if (resStart - resCount < 0)
      setBoolean(find("bsPrev"), "enabled", false);
    if (resStart + resCount < resNum)
      setBoolean(find("bsNext"), "enabled", true);      
    try {
      _showSearchPage(sTable);
    } catch (Exception e) {
      errorMsg(e.getMessage());
    }
  }
  
  public void nextPage(Object sTable) {
    int resNum = ((Integer)getProperty(sTable, "resNum")).intValue();
    resStart += resCount;
    if (resStart >= resNum) {
      resStart -= resCount;
      setBoolean(find("bsNext"), "enabled", false);
      return;
    }
    setBoolean(find("bsPrev"), "enabled", true);
    if (resStart + resCount >= resNum) {
      setBoolean(find("bsNext"), "enabled", false);
    }
    try {
      _showSearchPage(sTable);
    } catch (Exception e) {
      errorMsg(e.getMessage());
    }
  }
  
  private void _showSearchPage(Object sTable) throws Exception {
    removeAll(sTable);
    Hits hits = (Hits)getProperty(sTable, "hits");
    AllHitsCollector ahc = (AllHitsCollector)getProperty(sTable, "ahc");
    int resNum = (ahc == null) ? hits.length() : ahc.getSize();
    int max = Math.min(resNum, resStart + resCount);
    Object posLabel = find("resPos");
    setString(posLabel, "text", resStart + "-" + (max - 1));
    for (int i = resStart; i < max; i++) {
      int docid;
      float score;
      if (ahc != null) {
        docid = ahc.getDocId(i);
        score = ahc.getScore(i);
      } else {
        docid = hits.id(i);
        score = hits.score(i);
      }
      _createResultRow(i, docid, score, sTable);
    }
  }
  private void _createResultRow(int pos, int docId, float score, Object sTable) throws IOException {
    Object row = create("row");
    Object cell = create("cell");
    add(sTable, row);
    setString(cell, "text", String.valueOf(pos));
    setChoice(cell, "alignment", "right");
    add(row, cell);
    cell = create("cell");
    setString(cell, "text", String.valueOf(df.format(score)));
    setChoice(cell, "alignment", "right");
    add(row, cell);
    cell = create("cell");
    setString(cell, "text", String.valueOf(docId));
    setChoice(cell, "alignment", "right");
    add(row, cell);
    Document doc = ir.document(docId);
    putProperty(row, "docid", new Integer(docId));
    StringBuffer vals = new StringBuffer();
    for (int j = 0; j < idxFields.length; j++) {
      cell = create("cell");
      String[] values = doc.getValues(idxFields[j]);
      vals.setLength(0);
      if (values != null) for (int k = 0; k < values.length; k++) {
        if (k > 0) vals.append(' ');
        vals.append(escape(values[k]));
      }
      setString(cell, "text", vals.toString());
      add(row, cell);
    }
  }

  /**
   * Pop up a modal dialog explaining the selected result.
   * @param sTable Thinlet table widget containing selected search result.
   */
  public void explainResult(Object sTable) {
    Object row = getSelectedItem(sTable);
    if (row == null) return;
    Integer docid = (Integer) getProperty(row, "docid");
    if (docid == null) return;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    Query q = (Query) getProperty(sTable, "query");
    if (q == null) return;
    Object cbSimCust = find("cbSimCust");
    boolean useCustom = getBoolean(cbSimCust, "selected") && similarity != null;
    try {
      IndexSearcher is = new IndexSearcher(dir);
      if (useCustom) is.setSimilarity(similarity);
      Explanation expl = is.explain(q, docid.intValue());
      Object dialog = addComponent(null, "/xml/explain.xml", null, null);
      Object eTree = find(dialog, "eTree");
      addNode(eTree, expl);
      //setBoolean(eTree, "expand", true);
      add(dialog);
    } catch (Exception e) {
      e.printStackTrace();
      errorMsg(e.getMessage());
    }
  }

  private DecimalFormat df = new DecimalFormat("0.0000");

  private void addNode(Object tree, Explanation expl) {
    Object node = create("node");
    setString(node, "text", df.format((double) expl.getValue()) + "  " + expl.getDescription());
    add(tree, node);
    if (getClass(tree) == "tree") {
      setFont(node, getFont().deriveFont(Font.BOLD));
    }
    Explanation[] kids = expl.getDetails();
    if (kids != null && kids.length > 0) {
      for (int i = 0; i < kids.length; i++) {
        addNode(node, kids[i]);
      }
    }
  }

  public void gotoDoc(Object sTable) {
    Object row = getSelectedItem(sTable);
    if (row == null) return;
    Integer docid = (Integer) getProperty(row, "docid");
    if (docid == null) return;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    Document doc = null;
    try {
      doc = ir.document(docid.intValue());
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
      return;
    }
    _showDocFields(docid.intValue(), doc);
    Object tabpane = find("maintpane");
    setInteger(tabpane, "selected", 1);
    repaint();
  }

  private void _showTermDoc(Object fText, TermDocs td) {
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    try {
      Document doc = ir.document(td.doc());
      setString(find("docNum"), "text", String.valueOf(td.doc()));
      setString(find("tFreq"), "text", String.valueOf(td.freq()));
      _showDocFields(td.doc(), doc);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  public void deleteTermDoc(Object fText) {
    Term t = (Term) getProperty(fText, "term");
    if (t == null) return;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    if (readOnly) {
      showStatus(MSG_READONLY);
      return;
    }
    try {
      showNextTerm(find("fCombo"), fText);
      ir.deleteDocuments(t);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
    initOverview();
  }

  public void deleteDoc(Object docNum) {
    int docid = 0;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    if (readOnly) {
      showStatus(MSG_READONLY);
      return;
    }
    try {
      docid = Integer.parseInt(getString(docNum, "text"));
      ir.deleteDocument(docid);
      showDoc(docNum);
      initOverview();
      showStatus("Document #" + docid + " deleted OK.");
    } catch (Exception e) {
      showStatus(e.getMessage());
      e.printStackTrace();
    }

  }

  public void deleteDocList(Object sTable) {
    Object[] rows = getSelectedItems(sTable);
    if (rows == null || rows.length == 0) return;
    if (ir == null) {
      showStatus(MSG_NOINDEX);
      return;
    }
    if (readOnly) {
      showStatus(MSG_READONLY);
      return;
    }
    for (int i = 0; i < rows.length; i++) {
      Integer docId = (Integer) getProperty(rows[i], "docid");
      if (docId == null) continue;
      try {
        ir.deleteDocument(docId.intValue());
      } catch (Exception e) {
        continue;
      }
      remove(rows[i]);
    }
    initOverview();
  }

  public void actionAbout() {
    Object about = addComponent(this, "/xml/about.xml", null, null);
    Object lver = find(about, "lver");
    setString(lver, "text", "Lucene version: " + LucenePackage.get().getImplementationVersion());
  }

  /**
   * Pop up a modal font selection dialog.
   *
   */
  public void actionShowFonts() {
    Object sfonts = addComponent(this, "/xml/selfont.xml", null, null);
  }
  
  /**
   * Initialize the font selection dialog.
   * @param selfont font selection dialog
   */
  public void setupSelFont(Object selfont) {
    GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
    Font[] fonts = ge.getAllFonts();
    Object cbFonts = find(selfont, "fonts");
    String curfont = getFont().getFontName();
    float cursize = getFont().getSize2D();
    Object fsize = find(selfont, "fsize");
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(1);
    setString(fsize, "text", nf.format(cursize));
    removeAll(cbFonts);
    Object def = create("choice");
    setFont(def, "font", getFont().deriveFont(15.0f));
    setString(def, "text", curfont + " (default)");
    putProperty(def, "fnt", getFont());
    add(cbFonts, def);
    setInteger(cbFonts, "selected", 0);
    for (int i = 0; i < fonts.length; i++) {
      Object choice = create("choice");
      setFont(choice, "font", fonts[i].deriveFont(15.0f));
      setString(choice, "text", fonts[i].getFontName());
      putProperty(choice, "fnt", fonts[i]);
      add(cbFonts, choice);
      if (curfont.equalsIgnoreCase(fonts[i].getFontName()))
        setInteger(cbFonts, "selected", i + 1);
    }
  }
  
  /**
   * Show preview of the selected font.
   * @param selfont font selection dialog
   */
  public void selectFont(Object selfont) {
    Object preview = find(selfont, "fpreview");
    Object cbFonts = find(selfont, "fonts");
    Object fsize = find(selfont, "fsize");
    Font f = (Font)getProperty(getSelectedItem(cbFonts), "fnt");
    float size = getFont().getSize2D();
    try {
      size = Float.parseFloat(getString(fsize, "text"));
    } catch (Exception e) {
      e.printStackTrace();
    }
    f = f.deriveFont(size);
    Object[] items = getItems(preview);
    for (int i = 0; i < items.length; i++) {
      setPreviewFont(f, items[i]);
    }
  }
  
  private void setPreviewFont(Font f, Object item) {
    try {
      setFont(item, "font", f);
    } catch (IllegalArgumentException iae) {
      // shrug...
    }
    Object[] items = getItems(item);
    for (int i = 0; i < items.length; i++) {
      setPreviewFont(f, items[i]);
    }
  }
  
  /**
   * Set the default font in the UI.
   * @param selfont font selection dialog
   */
  public void actionSetFont(Object selfont) {
    Object cbFonts = find(selfont, "fonts");
    Object fsize = find(selfont, "fsize");
    Font f = (Font)getProperty(getSelectedItem(cbFonts), "fnt");
    float size = getFont().getSize2D();
    try {
      size = Float.parseFloat(getString(fsize, "text"));
    } catch (Exception e) {
      e.printStackTrace();
    }
    f = f.deriveFont(size);
    remove(selfont);
    setFont(f);
    courier = new Font("Courier", getFont().getStyle(), getFont().getSize());
    repaint();
  }
  
  /**
   * Returns current custom similarity implementation.
   * @return
   */
  public Similarity getCustomSimilarity() {
    return similarity;
  }
  
  /**
   * Set the current custom similarity implementation.
   * @param s
   */
  public void setCustomSimilarity(Similarity s) {
    similarity = s;
    Object cbSimCust = find("cbSimCust");
    Object cbSimDef = find("cbSimDef");
    Object simName = find("simName");
    if (similarity != null) {
      setString(simName, "text", similarity.getClass().getName());
      setBoolean(cbSimCust, "enabled", true);
    } else {
      setString(simName, "text", "");
      setBoolean(cbSimCust, "enabled", false);
      setBoolean(cbSimDef, "selected", true);
      setBoolean(cbSimCust, "selected", false);
    }
  }
  
  /**
   * Switch the view to display the SimilarityDesigner plugin, if present.
   *
   */
  public void actionDesignSimilarity() {
    LukePlugin designer = null;
    for (int i = 0; i < plugins.size(); i++) {
      if (plugins.get(i).getClass().getName().equals("org.getopt.luke.plugins.SimilarityDesignerPlugin")) {
        designer = (LukePlugin)plugins.get(i);
        break;
      }
    }
    if (designer == null) {
      showStatus("Designer Plugin not available");
      return;
    }
    // a bit tricky: plugins are put within panels, and these in tabs
    Object pluginsTab = find("pluginsTab");
    Object maintab = getParent(pluginsTab);
    int index = getIndex(maintab, pluginsTab);
    setInteger(maintab, "selected", index);
    Object pluginsTabs = find("pluginsTabs");
    Object tab = getParent(getParent(designer.getMyUi()));
    index = getIndex(pluginsTabs, tab);
    setInteger(pluginsTabs, "selected", index);
    repaint();
  }
  
  /**
   * Shut down Luke. If {@link #exitOnDestroy} is true (such as when Luke was
   * started from the main method), invoke also System.exit().
   */
  public boolean destroy() {
    if (ir != null) try {
      ir.close();
    } catch (Exception e) {}
    ;
    if (dir != null) try {
      dir.close();
    } catch (Exception e) {}
    ;
    try {
      Prefs.save();
    } catch (Exception e) {}
    ;
    if (exitOnDestroy) System.exit(0);
    return super.destroy();
  }

  public void actionExit() {
    destroy();
  }

  /**
   * Open URL in the system default browser.
   * @param url
   */
  public void goUrl(Object url) {
    String u = (String) getProperty(url, "url");
    if (u == null) return;
    try {
      BrowserLauncher.openURL(u);
    } catch (Exception e) {
      e.printStackTrace();
      showStatus(e.getMessage());
    }
  }

  /**
   * Start the GUI, and optionally open an index.
   * @param args index parameters
   * @return fully initialized Luke instance
   */
  public static Luke startLuke(String[] args) {
    Luke luke = new Luke();
    FrameLauncher f = new FrameLauncher("Luke - Lucene Index Toolbox, v 0.7.1 (2007-06-20)", luke, 700, 500);
    f.setIconImage(Toolkit.getDefaultToolkit().createImage(Luke.class.getResource("/img/luke.gif")));
    if (args.length > 0) {
      boolean force = false, mmap = false, ro = false;
      String pName = null;
      String script = null;
      for (int i = 0; i < args.length; i++) {
        if (args[i].equalsIgnoreCase("-ro")) ro = true;
        else if (args[i].equalsIgnoreCase("-force")) force = true;
        else if (args[i].equalsIgnoreCase("-mmap")) mmap = true;
        else if (args[i].equalsIgnoreCase("-index")) pName = args[++i];
        else if (args[i].equalsIgnoreCase("-script")) script = args[++i];
        else {
          System.err.println("Unknown argument: " + args[i]);
          usage();
          luke.actionExit();
          return null;
        }
      }
      if (pName != null) luke.openIndex(pName, force, mmap, ro);
      if (script != null) {
        LukePlugin plugin = luke.getPlugin("org.getopt.luke.plugins.ScriptingPlugin");
        if (plugin == null) {
          String msg = "ScriptingPlugin not present - cannot execute scripts.";
          System.err.println(msg);
          luke.actionExit();
        } else {
          ((ScriptingPlugin)plugin).execute("load('" + script + "');");
        }
      }
    } else luke.actionOpen();
    return luke;
  }
  
  /**
   * Main method. If you just want to instantiate Luke from other classes or scripts,
   * use {@link #startLuke(String[])} instead.
   * @param args
   */
  public static void main(String[] args) {
    exitOnDestroy = true;
    startLuke(args);
  }

  public static void usage() {
    System.err.println("Command-line usage:\n");
    System.err.println("Luke [-index path_to_index] [-ro] [-force] [-mmap] [-script filename]\n");
    System.err.println("\t-index path_to_index\topen this index");
    System.err.println("\t-ro\topen index read-only");
    System.err.println("\t-force\tforce unlock if the index is locked (use with caution)");
    System.err.println("\t-mmap\tuse MMapDirectory");
    System.err.println("\t-script filename\trun this script using the ScriptingPlugin.");
    System.err.println("\t\tIf an index name is specified, the index is open prior to");
    System.err.println("\t\tstarting the script. Note that you need to escape special");
    System.err.println("\t\tcharacters twice - first for shell and then for JavaScript.");
  }
  
  public static String escape(String[] values, String sep) {
    if (values == null) return null;
    if (values.length == 0) return "";
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < values.length; i++) {
      if (i > 0) sb.append(sep);
      sb.append(escape(values[i]));
    }
    return sb.toString();
  }
  
  public static String escape(String value) {
    if (value == null) return null;
    if (value.length() == 0) return value;
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < value.length(); i++) {
      char c = value.charAt(i);
      switch(c) {
        case '\t':
          sb.append("\\t");
          break;
        case '\r':
          sb.append("\\r");
          break;
        case '\n':
          sb.append("\\n");
        default:
          if (((int)c >= 32 && (int)c <= 127) || Character.isLetterOrDigit(c)) {
            sb.append(c);
          } else {
            sb.append("&#" + (int)c + ";");
          }
      }
    }
    return sb.toString();
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see java.awt.datatransfer.ClipboardOwner#lostOwnership(java.awt.datatransfer.Clipboard,
   *      java.awt.datatransfer.Transferable)
   */
  public void lostOwnership(Clipboard arg0, Transferable arg1) {

  }

}

/**
 * Simple Vector-like implementation of growable String array.
 * @author Andrzej Bialecki &lt;ab@getopt.org&gt;
 */
class GrowableStringArray {
  public int INITIAL_SIZE = 20;

  private int size = 0;

  private String[] array = null;

  public int size() {
    return size;
  }

  /**
   * Sets the value at specified index. If index is outside range the array is automatically
   * expanded.
   * @param index where to set the value
   * @param value
   */
  public void set(int index, String value) {
    if (array == null) array = new String[INITIAL_SIZE];
    if (array.length < index + 1) {
      String[] newArray = new String[index + INITIAL_SIZE];
      System.arraycopy(array, 0, newArray, 0, array.length);
      array = newArray;
    }
    if (index > size - 1) size = index + 1;
    array[index] = value;
  }

  /**
   * Appends the separator and value at specified index. If no value exists at the
   * specified position, this is equivalent to {@link #set(int, String)} - no separator
   * is appended in that case.
   * @param index selected position
   * @param sep separator
   * @param value value
   */
  public void append(int index, String sep, String value) {
    String oldVal = get(index);
    if (oldVal == null) {
      set(index, value);
    } else {
      set(index, oldVal + sep + value);
    }
  }

  /**
   * Return the value at specified index.
   * @param index
   * @return
   */
  public String get(int index) {
    if (array == null || index < 0 || index > array.length - 1) return null;
    return array[index];
  }
}